/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/


#include "RNAWithPairing.h"
#include "Utility.h"

extern ranlxs_state_t * vstate;	//global variable for random number generator 

/* static inline function */
static inline int min(int x, int y) {
	return (x > y) ? y : x;
}

template<class T>
static inline string NumToStrWidth(T x, int w){//T is is type parameter, T can be int,double, etc.
	std::ostringstream o;
	o.width(w);
	if (!(o << x)){
		cerr<<"Fatal error at file "<<__FILE__<<" line "<<__LINE__
			 <<" : Unable to convert "<<x<<" to string"<<endl; 
		exit(0);
	}
	return o.str();
}


/* constructors */
RNAWithPairing::RNAWithPairing(){}

RNAWithPairing::RNAWithPairing(const RNAWithPairing & r){
	
	s.reserve(r.s.size());
	p.reserve(r.p.size());
	w.reserve(r.w.size());
	c.reserve(r.c.size());

	s=r.s;
	p=r.p;
	w=r.w;
	c=r.c;
}


RNAWithPairing::RNAWithPairing(const vector<int> & _s, const vector<int> & _p){
	
	s.reserve(_s.size());
	p.reserve(_p.size());

	s = _s;
	p = _p;
}

/* return the sequence string. May add seq width */
string RNAWithPairing::getPairingStr(int width ) const {
	string pair;
	pair.reserve(p.size()-1);
	for(int i=0; i<p.size(); i++){
		if(p[i] == -1)
			pair.append(1, '.');
		else{
			p[i]>i ? pair.append(1, '(') : pair.append(1, ')');
			if(p[p[i]] != i){
				cerr<<"Fatal error at file "<<__FILE__<<" line "<<__LINE__
					<<" : pairing mismatch -- "
					<<i<<"=>"<<p[i]<<" vs "
					<<p[i]<<"=>"<<p[p[i]]<<endl;	
				exit(1);
			}
		}
	}
	if(width <= 0)
		return pair;
	else{
		string sout;
		sout.reserve(pair.size() * 2);
		int ind =0;
		while(ind < pair.size()){
			sout.append( pair.substr(ind, width) );
			sout.append(1, '\n');	
			ind += width;	
		}
		return sout;
	}
}

/* return the sequence string. May add seq width */
string RNAWithPairing::getSeqStr(int width ) const{ 
	
	string seq;
	int seqlen = s.size();	
	seq.reserve(seqlen);
	
	//translate numerical representation of RNA sequence into character representation
	for(int i=0; i<seqlen; i++){
		switch(s[i]){
			case 0: seq.append(1, 'A'); break;
			case 1: seq.append(1, 'U'); break;
			case 2: seq.append(1, 'C'); break;
			case 3: seq.append(1, 'G'); break;
			default: 
				cerr<<"Fatal error at file "<<__FILE__<<" line "<<__LINE__
					<<" : Unrecognized nucleotide designation: "<<s[i]<<endl; 
				exit(1); 
		}
	} 
	
	//return sequence
	if(width<=0)
		return seq;
	else{
		string sout;
		sout.reserve(seqlen*2);
		int ind =0;
		while(ind < seq.size()){
			sout.append( seq.substr(ind, width) );
			sout.append(1, '\n');
			ind += width;
		}
		return sout;
	}
}


string RNAWithPairing::repr(int rule, int rowwidth_in_tens) const{
	
	string sout("RNAWithPairing:");
	sout.reserve(5000);
	
	if(rule == 0){
		sout.append("\nseq    =");
		sout.append( getSeqStr() );
		sout.append("\npairing=");
		sout.append( getPairingStr() ); 
	}else{
		string seqstr = getSeqStr();
		string pairstr = getPairingStr();			
		sout.append("\n\n");
		
		int i = 0;
		int seqlen = seqstr.size();
		while(i<seqlen){
			int inext = min(seqlen, i + rowwidth_in_tens * 10);
			string thisstr = seqstr.substr(i, inext-i); //start at i, length inext-i
			string thisps = pairstr.substr(i, inext-i);
			int k = thisstr.size()/10; //use double = thisstr.size()/10.0 ?
			if(thisstr.size() % 10 >0)
				k++;
			for(int j= k-1; j>-1; j--){
				thisstr.insert(j*10, 1, ' ');	//insert(index, num, char)
				thisps.insert( j*10, 1, ' ');
			}
			
			//x.append("%5d %s\n"%(i,string.join(thisstr,sep="")))
            //x.append("      %s\n\n"%string.join(thisps,sep=""))

			sout.append(NumToStrWidth(i, 5));
			sout.append(1, ' ');
			sout.append(thisstr);
			sout.append(1, '\n');

			sout.append(6, ' ');
			sout.append(thisps);
			sout.append(2, '\n');

			i = inext;
		}
	}
	return sout;
}


string RNAWithPairing::terserepr() const{
	return  getSeqStr(60) + getPairingStr(60);	
}

string RNAWithPairing::verboserepr() const{
	return repr(1);
}

string RNAWithPairing::format(int resulttype) const{
	switch(resulttype){
		case 0:	return getSeqStr();
		case 1: return terserepr();
		case 2: return verboserepr();
		default: cerr<<"Fatal error at file "<<__FILE__<<" line "<<__LINE__
					<<" : Unrecognized resulttype: "<<resulttype<<endl; 
				exit(1);
	}
}



/* Create variation among site by assigning weight to each site follow 
   gamma distribution with mean 1 and variance 1/constR
*/
void RNAWithPairing::assignW(double gamma_shape, double gamma_rate, double gamma_maxweight){
	w.clear();
	int plen = p.size();
	w.reserve(plen);
//	if(type==1){
		for(int i=0; i<plen; i++){
			double weight = gsl_ran_gamma(vstate, gamma_shape, 1.0/gamma_rate);
			if(weight>gamma_maxweight)
				weight = gamma_maxweight;
			w.push_back(weight);
		}
//	}else{*/
//		for(int i=0; i<plen; i++)
//			w.push_back(1.0);
//	}
}

void RNAWithPairing::assignC(){
	c.clear();	
	int plen = p.size();	
	c.reserve(plen);
	for(int i=0; i<plen; i++){
		c.push_back(NumToStr(i));
	}
}


vector<int> * RNAWithPairing::getS(){
	return &s;
}

vector<int> * RNAWithPairing::getP(){
	return &p;
}

vector<double> * RNAWithPairing::getW(){
	return &w;
}

vector<string> * RNAWithPairing::getC(){
	return &c;
}

//END_OF_FILE
