/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/


/*******************************************************************************************
	A=0, U=1, C=2, G=3

	standard pairing:  0:AU  1:GU  2:GC  3:UA  4:UG  5:CG 

	compensatory mutation:
	{3,4,5},	//AU --> UA UG CG
	{3,4,5},	//GU --> UA UG CG
	{3,4,5},	//GC --> UA UG CG
	{0,1,2},	//UA --> AU GU GC 
	{0,1,2},	//UG --> AU GU GC
	{0,1,2},	//CG --> AU GU GC
********************************************************************************************/
#ifndef RNAWITHPAIRINGEVOLMODEL_H
#define RNAWITHPAIRINGEVOLMODEL_H

#include "Utility.h"
#include "Global.h"
#include "RNAWithPairing.h"

#include <cmath>
#include <vector>
#include <string>
#include <cstdio>
#include <fstream>
#include <iomanip>
#include <algorithm>

extern ranlxs_state_t * vstate;	//global variable for random number generator 
static const int _standardpairing[4][2]= {/* -1 added to avoid rugged array */
	{1, -1},  // A: U
	{0, 3},   // U: A, G
	{3, -1},  // C: G
	{1, 2}    // G: U, C
};


//pairing state:  0:AU  1:GU  2:GC  3:UA  4:UG  5:CG  6: all others
static const int _compensatory[6][3] = {/* -1 added to avoid rugged array */
	{3,4,5},	//AU --> UA UG CG
	{3,4,5},	//GU --> UA UG CG
	{3,4,5},	//GC --> UA UG CG
	{0,1,2},	//UA --> AU GU GC 
	{0,1,2},	//UG --> AU GU GC
	{0,1,2}		//CG --> AU GU GC
};


//A=0, U=1, C=2, G=3
static const int _pairs[6][2] ={
	{0,1},	//AU
	{3,1},	//GU
	{3,2},	//GC
	{1,0},	//UA
	{1,3},	//UG
	{2,3}	//CG
};

/* return an integer bounded between lb and ub (inclusive, i.e. [lb ub]) */
static inline int _randInt(int lb, int ub){	
	/* first generate a number between 0 .. N-1, then convert it to integer */
	double N = static_cast<double>(ub - lb  + 1);
	return static_cast<int>(N * ranlxs_get_double(vstate)) + lb;
}


/* return a random element from an integer set */
template<class T>
static inline T _getOneFromVector(vector<T> &s){
	return s[ _randInt(0,  s.size()-1) ];	//for index, get a random number in [0, 1,...size-1]	
}



/*	Event types: insertion, deletion, substitution, compensatory mutation*/
static const int constNumEvents=4;
static const double MAXENERGY = 1000000.0;

/*length distribution for insertion and deletion (power law)
	Based on reference below with scaling
	Reference: 
	[1] Empirical and structural models for insertions and deletions in the divergent evolution of proteins. 
		J Mol Biol. 1993 Feb 20;229(4):1065-82.  Benner SA, Cohen MA, Gonnet GH. 
	frequence of indel length x =  f(x) = m*x^(-1.7)
	trail truncated to limit indel length to 15
*/
static const int constIndelLimit = 6;
/*static const double _indellength[constIndelLimit]={
        0.5421, //probability of insert 1 nucleotide
        0.1669, //probability of insert 2 nucleotides
        0.0838, // ....
        0.0514,
        0.0351,
        0.0258,
        0.0198,
        0.0158,
        0.0129,
        0.0108,
        0.0092,
        0.0079,
        0.0069,
        0.0061,
        0.0054  //probability of insert nucleotide segment with maximal length 15
};*/

static const double _indellength[constIndelLimit]={
	0.59631, 	//probability of insert or delete 1 nucleotide
	0.18359,  	//probability of insert or delete 2 nucleotides
	0.09218,
	0.05654,
	0.03861,
	0.03277
};




/**************************************************************************************************
  Class Defintion
**************************************************************************************************/
class RNAWithPairingEvolModel{

private:
	
	/* parameters for the model */
	double N;		//population size
	double mu;		//mutation rate;
	double alpha;	//regularization factor in the energy based simulation schema, need to adjust according to Ne
	double kappa;	//ratio of indel to substitution
	int    distr_pool_size; /*number of RNAs used to construct the distribution of substitution rate*/
	double E_opt;
	double gamma_maxweight; /*maximal weight for a site*/
	bool    verbose;
	bool    fixed_global_DeltaE;
	double  global_DeltaE;
	
	/* variable for object being mutated: RNAWithPairing */
	RNAWithPairing *obj; //object being mutated, in this case, RNAWithPairing
	vector<int> pairedsite;	//list of paired sites
	vector<int> unpairedsite; //list of unpaired sites
		  
	/* in the schema, we compute minimal free energy and the various energy changes for root and some internal RNA.
	   these values are stored as static variable and used by later calls */
	static bool isEnergyComputed;	//indicating whether the minimal free energy has ever been computed.									
	static double deltaE;	//biggest energy difference between a mutant and the wildtype. deltaE<0; computed only once, i.e. only for the root RNA
	static int num_pooled_RNA; //number of pooled RNAs used to constructed the distribution of substitution rate
	static int NumMutatedRNA;  //number of mutated RNA seen so far

	double E_0;					//free energy of the wild type 
	
	vector<double> subRates;	//store all the rates for substutition
			
	void getSiteState();	
	double computeEnergy(const vector<int> * p, const vector<int> *s) const;
	double estimateR();
	double computeR();

	double insertTest(int mutsite, int mutnuc) const;
	double deleteTest(int mutsite) const;
	double multiinsertTest(int mutsite, vector<int> mutnucs)const;
	double multideleteTest(int mutsite, int indelsize) const;
	double substituteTest(int mutsite, int mutnuc) const;
	
	void insertCommit(int mutsite, int mutnuc);
	void deleteCommit(int mutsite);
	void multiinsertCommit(int mutsite, vector<int> mutnucs);
	void multideleteCommit(int mutsite, int indelsize);
	void substituteCommit(int mutsite, int mutnuc);
	
	int WeightedSampling() const;

	int getIndelLength();

public:
	
	RNAWithPairingEvolModel();
	void setparameters(double N, double mu, double E_opt, double kappa, int distr_pool_size, double gamma_maxweight, bool verbose);
	void setparameters(double N, double mu, double E_opt, double kappa, int distr_pool_size, double gamma_maxweight, bool verbose, bool _fixed_global_DeltaE, double _global_DeltaE);
	
	/* centralized function calling all mutation functions above */
	void mutate_exevent(RNAWithPairing * obj, double el, bool verboseflag, int nnodes);
};

#endif
//END_OF_FILE
