/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/
 

#include "RootedTree.h"
#include "Global.h"

const double MaxMinDIFF = 0.00001;	// used in isUltrametric to cope with machine error
									//If maxHeight/minHeight-1.0<MaxMinDIFF, consider it to be ultrametric tree

RootedTree::RootedTree(): root(NULL){}
RootedTree::~RootedTree(){ 		
	/* use tree traversal to delete all nodes */
	vector<RootedTreeNode *> n = getAllNodes();
	int len = n.size();
	for(int i = 0; i < len; i++){
		if(n[i] != NULL){
			delete n[i];
			n[i] = NULL;
		}
	}
	cout<<"Tree destructor called"<<endl;
}	

/* parse a string to get a tree structure, return root*/
void RootedTree::bystr(string & s){
	RootedTreeParser rtp(s);
	root = rtp.parse();
}

/* set the root node as given */
void RootedTree::setRoot(RootedTreeNode *r){
	root = r;
}

/* returns the root of the tree */
RootedTreeNode * RootedTree::getRoot() const{
	return root;
}

/* convert the tree into nexus format 
   print internal node labels if intnode is set to 1
*/
string RootedTree::nexus(int intnode) const {
	string sb;
	if(!root)
		return sb;
	sb = root->nexus(intnode);
	sb.append(1, ';');
	return sb;
}


/* contrat internal edges with zero edge length */
void RootedTree::contractZeroEdgeLength(){
	int len = (root->children).size();
	for(int i = 0; i < len; i++){
		root->children[i]->contractZeroEdgeLength();
	}
}


/* suppress node(s) with single children using preorder traversal (Single In Single Out) */
void RootedTree::suppressSISO(){
	
	stack<RootedTreeNode *> st;
	st.push(root);
	
	while( st.size() ){ //preorder traversal
		RootedTreeNode * v = st.top();
		st.pop();
		if( v->getChildren().size() == 1 && v->parent){	//SISO
			RootedTreeNode * vc = v->getChildren()[0];
			RootedTreeNode * vp = v->parent;
			vc->setEl(vc->getEl() + v->getEl());
			vp->insertChild(vc, vp->getChildIndex(v));
		}
		for(int i = v->children.size()-1; i> -1 ; i--)
			st.push(v->children[i]);
	}
}

/* make a copy of the tree */
RootedTree * RootedTree::copy() const {
	RootedTree * tgt = new RootedTree();
	assert(tgt != 0);
	tgt->setRoot(root->copy());
	return tgt;
}


/* test if the tree is binary using nonrecursive dfs */
bool RootedTree::isBinary() const {

	stack<RootedTreeNode *> st;
	st.push(root);
	
	while( st.size() ){ //preorder traversal
		RootedTreeNode * v = st.top();
		st.pop();
		
		if(v->children.size()>2 || v->children.size()==1)//leaf node has no child
			return false;
		for(int i = v->children.size()-1; i> -1 ; i--)
			st.push(v->children[i]);
	}	
	return true;
}


/* test if a tree is ultrametric */
bool RootedTree::isUltrametic() const{

	double maxh = maxHeight();
	double minh = minHeight();

	//if(maxh-minh < MaxMinDIFF)	//MaxMinDIFF = 0.001
	if(maxh/minh -1.0 <MaxMinDIFF)
		return true;
	else
		return false;
}

/* get the maximal distance from a leaf to root */
double RootedTree::maxHeight() const{
	vector<RootedTreeNode *> ls = leafset();
	double mh = 0.0;
	double h;
	RootedTreeNode * temp;
	for(unsigned int i=0; i<ls.size(); i++){
		h = 0.0;
		temp = ls[i];
		while(temp != root){
			h += temp->getEl();
			temp = temp->getParent();
		}
		if(i==0) mh =  h;
		else if(h>mh) mh = h;
	}
	return mh;
}


/* get the minimal distance from a leaf to root */
double RootedTree::minHeight() const{
	vector<RootedTreeNode *> ls = leafset();
	double mh = 0.0;
	double h;
	RootedTreeNode * temp;
	for(unsigned int i=0; i<ls.size(); i++){
		h = 0.0;
		temp = ls[i];
		while(temp != root){
			h += temp->getEl();
			temp = temp->getParent();
		}
		if(i==0) mh =  h;
		else if(h<mh) mh = h;		
	}

    return mh;
}


/* retrieve the set of leaf nodes using preorder traversal */
vector<RootedTreeNode *> RootedTree::leafset()const {
	stack<RootedTreeNode *> st;
	st.push(root);

	vector<RootedTreeNode *>  ls;
	while( st.size() ){ //preorder traversal
		RootedTreeNode * v = st.top();
		st.pop();
		
		if(v->isLeaf())
			ls.push_back(v);
		for(int i = v->children.size()-1; i> -1 ; i--)
			st.push(v->children[i]);	
	}	
	
	return ls;	
}

/* retrieve the set of all nodes (internal and leaf nodes) using preorder traversal */
vector<RootedTreeNode *> RootedTree::getAllNodes()const {
	stack<RootedTreeNode *> st;
	st.push(root);

	vector<RootedTreeNode *>  ls;
	while( st.size() ){ //preorder traversal
		RootedTreeNode * v = st.top();
		st.pop();
		
		ls.push_back(v);	//store all nodes
		for(int i = v->children.size()-1; i> -1 ; i--)
			st.push(v->children[i]);	
	}	
	
	return ls;	
}


/* scale tree edge by sf */
void RootedTree::scale(double sf){
	stack<RootedTreeNode *> st;
	st.push(root);

	int i=0;
	while( st.size() ){ //preorder traversal
		i++;
		RootedTreeNode * v = st.top();
		st.pop();
		
		if(i!=1)
			v->setEl(v->getEl()*sf);	//scale edge length

		for(int i = v->children.size()-1; i> -1 ; i--)
			st.push(v->children[i]);	
	}	
}


/* retrieve the circular order of the leaves using preorder traversal */
vector<RootedTreeNode *>  RootedTree::circularorder() const{
	return leafset();
}

/* Retrieve the list of labels given the list of nodes */
vector<string> RootedTree::nodelist_to_labellist(vector<RootedTreeNode *> nodelist) const{
	
	vector<string> y;
	for(unsigned int i = 0; i < nodelist.size(); i++)
		y.push_back(nodelist[i]->getLabel());
	return y;
}


//END_OF_FILE
