/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/


#include "RootedTreeNode.h"

using namespace std;

/***************************************************************************************************************
 Definition of class RootedTreeNode 
***************************************************************************************************************/

/* global constant variable */
const double FLOAT_ZERO_BOUND =  0.00000001;	//a small number used to determine if another number is close to zero


/* static inline function */
static inline int min(int x, int y) { return (x <y) ? x : y; }

template<typename T> 
static inline string NumToStr(T x){//T is is type parameter, T can be int, double
	std::ostringstream o;
	//if (!(o << x)){
	if (!(o << setiosflags(ios::fixed | ios::showpoint)<< setprecision(8)<< x)){
		cerr<<"Fatal error at file "<<__FILE__<<" line "<<__LINE__
			<<" : Unable to convert "<<x<<" to string"<<endl; 
		exit(0);
	}	
	return o.str();
}


/* Initialize static data member at file scope */
int RootedTreeNode::nodecnt = 0;
int RootedTreeNode::getNodecnt(){ return nodecnt;}
  
/* default constructor for new node */
RootedTreeNode::RootedTreeNode( ){	
	parent = NULL; 
	el = 1.0;
	elFlag = false;		

	serial = nodecnt + 1;
	++nodecnt;
}


/* utility function, used internally */
void RootedTreeNode::_setParent( RootedTreeNode * p){
	parent = p;

}
RootedTreeNode * RootedTreeNode::getParent() const{	
	return parent;
}

int RootedTreeNode::getSerial() const{
	return serial;
}
/* return pointers to children nodes */
vector<RootedTreeNode *> RootedTreeNode::getChildren() const {
	return children;
}

/* add a child node to this node, set child node's parent to this node */
void RootedTreeNode::addChild(RootedTreeNode * c){
	children.push_back(c);
	c->_setParent(this);
}

/* insert node as the (i+1)'th child of this node. i.e. having index i; index starts at 0;
   If i is greater than the number of children of this node minus one, add c as the last child.
*/
void RootedTreeNode::insertChild(RootedTreeNode * c,  unsigned int i){
	if(i>=children.size())
		children.push_back(c);
	else{
		vector<RootedTreeNode *>::iterator theIterator = children.begin(); // Iterator == children.begin() = 0;
		for(unsigned int k=0; k<i; k++)	//vector index starts at zero
			theIterator++;
		children.insert(theIterator, c);	//insert c BEFORE Iterator !!!	
	}
	
	c->_setParent(this);
}


/* remove a child, internal use only*/
void RootedTreeNode::_removeChild(RootedTreeNode * c){
	vector<RootedTreeNode *>::iterator theIterator = children.begin(); // Iterator == children.begin() = 0;
	unsigned int i;
	for(i = 0; i < children.size(); i++){	//vector index starts at zero
		if(children[i] != c)
			theIterator++;
		else
			break;
	}		
	delete children[i];	//deletion will NOT propogates to children[i]'s children, so only children[i] is removed
	children.erase(theIterator);	
}



/* get index of child node in the list of children */
int RootedTreeNode::getChildIndex(const RootedTreeNode * c) const{
	unsigned int i = 0;
	while(i<children.size())			//index starts from zero
		if(children[i] == c)
			return i;
	return -1;
}

/* return true if the node is a leaf*/
bool RootedTreeNode::isLeaf() const{
	return (children.size()>0)? false : true;
}

/* return true if the upedge is associated with edge length */
bool RootedTreeNode::hasEl() const {
	return elFlag;
}

/* return the edge length */
double RootedTreeNode::getEl() const{
	return el;	     
}
/* set edge length between the node and its parent to be 1 */
void RootedTreeNode::setEl(double e){
	el = e;
	elFlag = true;
}

/*set population size */
void RootedTreeNode::setN(long n){
	N = n;	
}

long RootedTreeNode::getN() const{
	return N;
}

/* return the label string of the node */
string RootedTreeNode::getLabel() const{
	
	//return label;	 

	if(isLeaf())
		return label;	 
	else{
		if(label.size()<1){
			string r("_I");	
			r.append(NumToStr(serial));	//use inline function
			return r;
		}
		else
			return label;
	}
}

/* set node label */
void RootedTreeNode::setLabel(const string & l){
	label = l;	
}



/*manipulation function*/
/* make a copy of the subtree rooted at this node.
 * create a rooted tree node (together with the nodes in the subtree rooted by this) as a clone of the input node
*/
RootedTreeNode * RootedTreeNode::copy() const{
	RootedTreeNode * tgt = new RootedTreeNode();
	assert(tgt != 0);
	tgt->setEl(el);
	if(isLeaf( ))
		tgt->label = label;
	else{
		for(unsigned int i = 0; i < children.size(); i++)
			tgt->addChild(children[i]->copy());
	}
	return tgt;
}


/* convert the subtree rooted at this node to nexus format. 
 * Print internal node labels if intnode is set to 1 */
string RootedTreeNode::nexus(int intnode) const{
	string sb;
	if(isLeaf()){
		sb.append(getLabel());
		if(hasEl()){
			sb.append(1, ':');
			sb.append(NumToStr(el));
		}
	}else{
		sb.append(1, '(');
		int nc = children.size();
		for(int i = 0; i < nc; i++){
			sb.append(children[i]->nexus(intnode));
			if(i < nc-1)
				sb.append(1, ',');
		}
		sb.append(1, ')' );
		if(hasEl()){
			if(intnode == 1)
				sb.append( getLabel() );
			sb.append(1, ':');
			sb.append(NumToStr(el));
		}
	}	
	return sb;
}

/* apply contraction to every edge that has zero edge length */
void RootedTreeNode::contractZeroEdgeLength(){
	if(isLeaf())
		return;
	
	int nc = children.size();
	for(int i = 0; i < nc; i++)
		children[i]->contractZeroEdgeLength();
	
	if(el < FLOAT_ZERO_BOUND){ //WARNING: original python code use if(el == 0.0), determine if a floating number is zero can be tricky, see above		
		int  ind = parent->getChildIndex(this); 
		for(unsigned int i =0; i < children.size(); i++){ //add this node?s children as its parent?s children
			parent->insertChild(children[i], ind++);
		}
		//remove this node from its parent's children
		parent->_removeChild(this);  //WARNING: might be problematic
	}	
}

//END_OF_FILE
