/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/
 

/*******************************************************************************************
	Rooted tree node with pointers to parent and children.			
********************************************************************************************/
#ifndef ROOTEDTREENODE_H
#define ROOTEDTREENODE_H

#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <cassert>
#include <iomanip>


using namespace std;

class RootedTreeNode {

	friend class RootedTree;

private:

	static int nodecnt;					//num of RootedTreeNode instantiated and still alive, can be decrease by destructor 
	int serial;							//unique serial number for this node. generated by constructor and not mutatable 
	RootedTreeNode * parent;			//pointer to its parent
	vector<RootedTreeNode * > children;	//pointers to its children
	string label;						//leaf label
	bool elFlag;						//flag showing it has edge length
	double el;							//edge length
	long N;								//population size 
	
	
	/* utility function, used internally */
	void _setParent( RootedTreeNode * );		

	//remove a child (i.e. the subtree rooted at this child is pruned
	void _removeChild(RootedTreeNode * c);		


public:	
	/*constructor and destructor*/
	//default construtor
	explicit RootedTreeNode();	//use 'explicit' keyword to disable automatic type conversions -- generally a good idea
	
	/*	copy constructor		
		RootedTreeNode(const RootedTreeNode & );

		constructor
		RootedTreeNode(RootedTreeNode *, vector<RootedTreeNode *>, string, bool, double);

		destructor
		~RootedTreeNode();							
	*/
	
	/*overloaded assignment operator. const return avoids: (a1=a2) = a3; */
	const RootedTreeNode &operator= ( const RootedTreeNode &);	
	
	/*get functions*/
	//return the serial number of the node
	int getSerial() const;

	//return pointer to parent node
	RootedTreeNode * getParent() const;

	//return pointers to children nodes
	vector<RootedTreeNode *> getChildren() const;
	
	//return index of a child node in children vector
	int getChildIndex(const RootedTreeNode *) const; 
	
	//return node label
	string getLabel() const;

	//return the edge length between the node and its parent
	double getEl() const;
	
	//return population size associated with this node
	long getN()	const; 
	
	/*set functions*/
	//add a child node to this node, set child node's parent to this node
	void addChild(RootedTreeNode *);
	
	//insert node cas the i'th child of this node. If i is greater than the number of children of this node plus one, add c as the last child.
	void insertChild(RootedTreeNode * c, unsigned int i);

	//set node label
	void setLabel(const string &);
	
	//set edge length between the node and its parent 
	void setEl(double = 1.0); 

	//set population size
	void setN(long);
	
	/*boolean function*/
	bool isLeaf() const;
	bool hasEl() const;
	
	/*manipulation function*/
	//make a copy of the subtree rooted at this node. create a rooted tree node (together with the nodes in the subtree rooted by this) as a clone of the input node.
	RootedTreeNode * copy() const; 

	//convert the subtree rooted at this node to nexus format. Print internal node labels if intnode is set to 1
	string nexus(int intnode = 0) const;	
       
	//apply contraction to evry edge that has zero edge length
	void contractZeroEdgeLength();

	//Depth-first traversal
	//f is an object such that f.method(node) can be called
	//Set options: preorder = 0; postorder = 1; inorder = 2; Eulerian = 3 (preorder and postorder)
	//void dfs(int options, RootedTreeNode & f) const;	

	/*static function*/
	//return static variable
	static int getNodecnt();

};
#endif
//END_OF_FILE
