/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/


#include "RootedTreeParser.h"

using namespace std;

/* convert string to double	*/
static inline double StrToDouble(const string & s){ 
	return atof( s.c_str() );	
}

/* default constructor */
RootedTreeParser::RootedTreeParser() : head(0){}

/* constructor */
RootedTreeParser::RootedTreeParser(const string &s): head(0), treestr(s){}

/* read a label */
string RootedTreeParser::readLabel() {
	string sb;
	sb.reserve(60);
	char c = treestr[head];
	int len = treestr.size();
	
	while( isalnum(c) || c == '_' || c == '-' ){	
		sb.append(1, c);
		head++;	
		if(head >= len)
			break;
		c = treestr[head];
	} 
	return sb;
}

/* read a numerical value */
double RootedTreeParser::readVal(){
	string sb;
	char c = treestr[head];
	int len = treestr.size();
	
	while(isdigit(c) || c == '.' || c == 'e' || c =='E' || c == '+' || c == '-'  ){	
		sb.append(1, c);
		head++;	
		if(head >= len)
			break;
		c = treestr[head];
	} 
	return StrToDouble(sb);	//convert string to double	
}

/* Parse the internal string (tree in nexus format);
 * return the tree root as an object of the RootedTreeNode class
 * rturn 0 if the internal string is not a legitimate nexus format;
 * should be called by an RootedTree object */
RootedTreeNode * RootedTreeParser::parse(){
	
	RootedTreeNode * root = NULL;
	RootedTreeNode * currentnode = NULL;
	RootedTreeNode * currentparent = NULL;
	//RootedTreeNode * currentlastsibl = NULL;
	RootedTreeNode * tempnode = NULL;

	int headlast = 0;
	
	head = 0;	//set head to beginning

	int len = treestr.size();
	
	/* check if parenthesis is matching */
	int i = 0;
	int c = 0;
	while(i<len){
		if( treestr[i] == '(' )
			c++;
		else if(treestr[i] == ')')
			c--;
		i++;
	}
	if(c!= 0){
		cerr<<"Error: unbalanced parenthesis in the input tree"<<endl;
		exit(1);
	}


	while(head < len){

		headlast = head;

		if( treestr[head] == '(' ){ //new node
			tempnode = new RootedTreeNode();
			assert(tempnode != 0);
			tempnode->setEl(1.0);	//default edge length, incase inpute does not havd edge length
			if( root == NULL){	// this is  root			
				root = tempnode; 
				root->setEl(0.0);
				currentparent = NULL;
			}else{
				if( treestr[head-1] == '(' ) //first child
					currentnode->addChild( tempnode );
				else{
					currentparent->addChild( tempnode );
					currentparent = currentnode;		//new level
				}
			}
			currentnode = tempnode;
			head++;
			continue;	//terminate here and start next iteration in the while loop
		}		

		if( treestr[head] == ')' ){ //end list of children
			if(currentnode == NULL ){
				cerr<<"Unbalance node found in nexus format at position "<<head<<endl;
				return 0;	//unbalanced node			
			}
			currentnode = currentnode->getParent();	//up one level
			currentparent = currentnode->getParent();
			++head;
			continue;
		}

		if(treestr[head] == ':'){	//starts edge length
			++head;
			currentnode->setEl(readVal());
			continue;
		}

		if(treestr[head] == ';'){	//end of tree string
			++head;
			break;	
		}

		if(treestr[head] == ','){	//next child
			++head;
			continue;
		}

		if(isalnum(treestr[head]) || treestr[head] == '-' || treestr[head] == '_'){	//a label
			
			if(treestr[headlast-1] == ')'){//internal node which is already created when reading '('
				currentnode->setLabel(readLabel());
				continue;
			}

			tempnode = new RootedTreeNode();
			assert(tempnode != 0);
			tempnode->setLabel(readLabel());

			if(headlast == 0){ //one leaf
				currentnode = tempnode;
				root = currentnode;
				continue;
			}
			
			if(treestr[headlast - 1] == '('){
				currentnode->addChild(tempnode);
				currentparent = currentnode;
			}else{
				currentparent->addChild(tempnode);
			}



			currentnode = tempnode;
			continue;
		}
		
		if(treestr[head]==' '){
			head++;
			continue;
		}

		
		cerr<<"Error: char "<<treestr[head]<<" at "<<head<<" in the input tree"<<endl;
		exit(1);
		++head;
	}	
	
	return root;
}
//END_OF_FILE
