/* Copyright (c) 2008, Sheng Guo, Li-San Wang and Junhyong Kim, University of
  Pennsylvania.  All Rights Reserved.
 
  You may not use this file except in compliance with the License
  located in the top directory of this distribution.  You may obtain
  a copy of the License at
 
      http://kim.bio.upenn.edu/software/index.shtml
 
  Unless required by applicable law or agreed to in writing, this
  software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
  CONDITIONS OF ANY KIND, either express or implied.  See the License
  for the specific language governing permissions and limitations
  under the License. 
*/


#include "RootedTreeNode.h"
#include "RootedTree.h"
#include "RootedTreeParser.h"
#include "RNAWithPairing.h"
#include "RNAWithPairingParser.h"
#include "RNAWithPairingEvolModel.h"
#include "EvolutionTree.h"
#include "MutateOnTree.h"
#include "Parameters.h"
#include "Utility.h"
#include "Global.h"

#include <iostream>
#include <fstream>
#include <string>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <getopt.h>
#include <unistd.h>



using namespace std;
ranlxs_state_t * vstate;	//global variable for random number generator 

void	display_tree_info(RootedTree * rt);

int main(int argc, char* argv[]){		
		
		vstate = new ranlxs_state_t;
		ranlxs_set_lux(vstate, static_cast<long>(time(0)), 373);	//initialize; second para is the seed;

		/*************************************************************
		     Check and initialize parameters
		/*************************************************************/
		Parameters Para;
		if(!Para.process_parameters(argc, argv)){
			delete vstate;
			exit(EXIT_FAILURE);
		}
		Para.display_parameters();
	

		/*************************************************************
			Initialize Evolution Tree
		*************************************************************/
		/* read and scale tree */
		string treestr;
		ifstream inTree(Para.in_tre, ios::in);
		do{
			getline(inTree, treestr);
			removeLeadTrailWS(treestr);
		}while( treestr.length()==0 );	//get first non-empty line as tree string
		inTree.close();
		
		RootedTree * rt = new RootedTree();
		rt->bystr(treestr);		
		rt->scale(Para.bsf);		
		display_tree_info(rt);

		EvolutionTree * evtree = new EvolutionTree(rt);
			
		
		/*************************************************************
			Initialize RNA structure
		*************************************************************/
		/* read RNA secondary structure */
		string rootrnaseq;	
		string root2sstr;
		string tempstr;
		int step = 0;
		ifstream inRoot(Para.in_rna, ios::in);
		while(getline(inRoot, tempstr)){//discard newline char
			removeLeadTrailWS(tempstr);	//remove leading and trailing white space			
			if( tempstr.size() == 0 )
				continue;
			if(step == 0 && tempstr[0] != '>')
				step = 1;
			else if(step == 1 && (tempstr[0] =='<' || tempstr[0] =='>' || tempstr[0] =='(' || tempstr[0] ==')' || tempstr[0] =='.'))
				step = 2;
			
			if(step == 1)
				rootrnaseq.append(tempstr);
			else if(step ==2)
				root2sstr.append(tempstr);
		}
		inRoot.close();
		strReplace(rootrnaseq, 'T', 'U'); //replace 'T' by 'U'
		strReplace(rootrnaseq, 't', 'u'); //replace 't' by 'u'

		RNAWithPairingParser * rp = new RNAWithPairingParser();
		RNAWithPairing * rootrna = rp->parse(rootrnaseq, root2sstr);
		delete rp;

		/********************************************************
		create rate variation-among-site of the RNA sequence
		********************************************************/
		rootrna->assignW(Para.gamma_shape,Para.gamma_rate, Para.gamma_maxweight);
		rootrna->assignC();
		cout<<"starting simulation"<<endl;

		/********************************************************
			Mutation
		********************************************************/
		RNAWithPairingEvolModel * rnaevolmodel = new RNAWithPairingEvolModel();	
		rnaevolmodel->setparameters(Para.N, Para.alpha, Para.E_opt, 1.0/Para.kappa, Para.distr_pool_size, Para.gamma_maxweight, Para.verbose, Para.fixed_global_DeltaE, Para.global_DeltaE);
		
		MutateOnTree * mm = new MutateOnTree(evtree,rnaevolmodel,rootrna, Para.verbose);	

		remove(Para.out_nuc.c_str());
		remove(Para.out_sec.c_str());
		remove(Para.out_sid.c_str());
		remove(Para.out_pos.c_str());
		remove(Para.out_tre.c_str());


		time_t rawtime;
		struct tm * timeinfo;
		time ( &rawtime );
		timeinfo = localtime ( &rawtime );
		
		mm->mutate(Para.out_nuc.c_str(), Para.out_pos.c_str(), Para.out_sec.c_str(), Para.out_sid.c_str(), Para.out_tre.c_str(), Para.verbose);

		/* reclaim memory  Note: delete rootrna is not needed. the deletion is done in Evolution.h EvolutionTree.eraseObj()	*/					
		delete vstate;
		delete evtree;	
		delete rnaevolmodel;
		delete mm; 	

		return EXIT_SUCCESS;
}


void	display_tree_info(RootedTree * rt){

	cout<<"This is a ";
	rt->isBinary() ? cout<<"binary " : cout<<"non-binary ";
	rt->isUltrametic() ? cout<<"ultrametric" : cout<<"non-ultrametric";
	cout<<" tree"<<endl;
	
	cout<<"maximal distance from leaves to root is "<<rt->maxHeight()<<endl;
	cout<<"minimal distance from a leaf to root is "<<rt->minHeight()<<endl;		
}
